from spb_curate.curate.model import annotation_types
from spb_curate.curate.model.annotation_types import (
    BoundingBox,
    Category,
    Contour,
    Cuboid2D,
    Edge,
    Face,
    Keypoint,
    Keypoints,
    KeypointState,
    Path,
    Point,
    Polygon,
    Polyline,
    RotatedBox,
)
